@@ -2,9 +2,8 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.pool import Pool
-from . import user
-from . import party
-from . import ir
+
+from . import ir, party, user
 
 
 def register():
