@@ -9,6 +9,7 @@ import time
 import urllib.parse
 import warnings
 from email.header import Header
+from secrets import token_hex
 
 try:
     import bcrypt
@@ -25,17 +26,20 @@ from sql.operators import Equal
 
 from trytond.config import config
 from trytond.exceptions import RateLimitException
+from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, DeactivableMixin, fields, Unique, Exclude,
-    avatar_mixin)
+    DeactivableMixin, Exclude, Index, ModelSQL, ModelView, Unique,
+    avatar_mixin, fields)
 from trytond.pool import Pool
 from trytond.pyson import Eval
 from trytond.report import Report, get_email
-from trytond.transaction import Transaction
+from trytond.res.user import CRYPT_CONTEXT, LoginAttempt
 from trytond.sendmail import sendmail_transactional
+from trytond.tools.email_ import (
+    EmailNotValidError, normalize_email, set_from_header, validate_email)
+from trytond.transaction import Transaction
 
-from trytond.ir.session import token_hex
-from trytond.res.user import LoginAttempt, CRYPT_CONTEXT
+from .exceptions import UserValidationError
 
 logger = logging.getLogger(__name__)
 
@@ -44,7 +48,7 @@ def _send_email(from_, users, email_func):
     from_cfg = config.get('email', 'from')
     for user in users:
         msg, title = email_func(user)
-        msg['From'] = from_ or from_cfg
+        set_from_header(msg, from_cfg, from_ or from_cfg)
         msg['To'] = user.email
         msg['Subject'] = Header(title, 'utf-8')
         sendmail_transactional(from_cfg, [user.email], msg)
@@ -69,17 +73,17 @@ class User(avatar_mixin(100), DeactivableMixin, ModelSQL, ModelView):
     __name__ = 'web.user'
     _rec_name = 'email'
 
-    email = fields.Char('E-mail', select=True,
+    email = fields.Char(
+        "E-mail",
         states={
             'required': Eval('active', True),
-            },
-        depends=['active'])
+            })
     email_valid = fields.Boolean('E-mail Valid')
-    email_token = fields.Char('E-mail Token', select=True)
+    email_token = fields.Char("E-mail Token", strip=False)
     password_hash = fields.Char('Password Hash')
     password = fields.Function(
         fields.Char('Password'), 'get_password', setter='set_password')
-    reset_password_token = fields.Char('Reset Password Token', select=True)
+    reset_password_token = fields.Char("Reset Password Token", strip=False)
     reset_password_token_expire = fields.Timestamp(
         'Reset Password Token Expire')
     party = fields.Many2One('party.party', 'Party', ondelete='RESTRICT')
@@ -96,6 +100,10 @@ class User(avatar_mixin(100), DeactivableMixin, ModelSQL, ModelView):
                     where=table.active == Literal(True)),
                 'web_user.msg_user_email_unique'),
             ]
+        cls._sql_indexes.update({
+                Index(table, (table.email, Index.Equality())),
+                Index(table, (table.email_token, Index.Equality())),
+                })
         cls._buttons.update({
                 'validate_email': {
                     'readonly': Eval('email_valid', False),
@@ -107,18 +115,6 @@ class User(avatar_mixin(100), DeactivableMixin, ModelSQL, ModelView):
                     },
                 })
 
-    @classmethod
-    def __register__(cls, module_name):
-        super(User, cls).__register__(module_name)
-
-        table_h = cls.__table_handler__(module_name)
-
-        # Migration from 4.6
-        table_h.not_null_action('email', 'remove')
-
-        # Migration from 4.6: replace unique by exclude
-        table_h.drop_constraint('email_unique')
-
     @classmethod
     def default_email_valid(cls):
         return False
@@ -144,10 +140,15 @@ class User(avatar_mixin(100), DeactivableMixin, ModelSQL, ModelView):
                         }])
         cls.write(*to_write)
 
+    @fields.depends('party', 'email')
+    def on_change_party(self):
+        if not self.email and self.party:
+            self.email = self.party.email
+
     @classmethod
     def _format_email(cls, users):
         for user in users:
-            email = user.email.lower()
+            email = normalize_email(user.email).lower()
             if email != user.email:
                 user.email = email
         cls.save(users)
@@ -171,6 +172,26 @@ class User(avatar_mixin(100), DeactivableMixin, ModelSQL, ModelView):
         users = sum(args[0:None:2], [])
         cls._format_email(users)
 
+    @classmethod
+    def validate_fields(cls, users, fields_names):
+        super().validate_fields(users, fields_names)
+        cls.check_valid_email(users, fields_names)
+
+    @classmethod
+    def check_valid_email(cls, users, fields_names=None):
+        if fields_names and 'email' not in fields_names:
+            return
+        for user in users:
+            if user.email:
+                try:
+                    validate_email(user.email)
+                except EmailNotValidError as e:
+                    raise UserValidationError(gettext(
+                            'web_user.msg_user_email_invalid',
+                            user=user.rec_name,
+                            email=user.email),
+                        str(e)) from e
+
     @classmethod
     def authenticate(cls, email, password):
         pool = Pool()
@@ -407,7 +428,7 @@ class User_PartySecondary(ModelSQL):
     __name__ = 'web.user-party.party.secondary'
 
     user = fields.Many2One(
-        'web.user', "User", required=True, select=True, ondelete='CASCADE')
+        'web.user', "User", required=True, ondelete='CASCADE')
     party = fields.Many2One(
         'party.party', "Party", required=True, ondelete='CASCADE')
 
@@ -423,9 +444,9 @@ class UserSession(ModelSQL):
     __name__ = 'web.user.session'
     _rec_name = 'key'
 
-    key = fields.Char('Key', required=True, select=True)
+    key = fields.Char("Key", required=True, strip=False)
     user = fields.Many2One(
-        'web.user', 'User', required=True, select=True, ondelete='CASCADE')
+        'web.user', "User", required=True, ondelete='CASCADE')
 
     @classmethod
     def __setup__(cls):
@@ -436,6 +457,13 @@ class UserSession(ModelSQL):
             ('key_unique', Unique(table, table.key),
                 'web_user.msg_user_session_key_unique'),
             ]
+        cls._sql_indexes.update({
+                Index(
+                    table,
+                    (Coalesce(table.write_date, table.create_date),
+                        Index.Range())),
+                Index(table, (table.key, Index.Equality())),
+                })
 
     @classmethod
     def default_key(cls, nbytes=None):
@@ -449,7 +477,7 @@ class UserSession(ModelSQL):
         cursor.execute(*table.delete(
                 where=(
                     Coalesce(table.write_date, table.create_date)
-                    - CurrentTimestamp()) > cls.timeout()))
+                    < CurrentTimestamp() - cls.timeout())))
 
         session = cls(user=user)
         session.save()
@@ -521,4 +549,11 @@ class EmailResetPassword(Report):
     def get_context(cls, records, header, data):
         context = super().get_context(records, header, data)
         context['extract_params'] = _extract_params
+        expire_delay = (
+            records[0].reset_password_token_expire - datetime.datetime.now())
+        # Use a precision of minutes
+        expire_delay = datetime.timedelta(
+            days=expire_delay.days,
+            minutes=round(expire_delay.seconds / 60))
+        context['expire_delay'] = expire_delay
         return context
